.v-expansion-panel-content .row[data-v-0b22f425]:first-child{margin-top:12px}.disabled-pointer[data-v-0b22f425]{pointer-events:none}.v-expansion-panel-content .row[data-v-309e10ce]:first-child{margin-top:12px}.v-textarea-json[data-v-309e10ce]{font-family:Monaco,Menlo,Consolas,Bitstream Vera Sans Mono,monospace}