#
# voice-skill-sdk
#
# (C) 2021, Deutsche Telekom AG
#
# Deutsche Telekom AG and all other contributors /
# copyright owners license this file to you under the MIT
# License (the "License"); you may not use this file
# except in compliance with the License.
# You may obtain a copy of the License at
#
# https://opensource.org/licenses/MIT
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

"""Magenta Voice Skill SDK: version info"""

__name__ = "skill-sdk"
__description__ = "Magenta Voice Skill SDK for Python"
__url__ = "https://github.com/telekom/voice-skill-sdk/"
__version__ = "1.4.0"
__author__ = "Deutsche Telekom Voicification Suite"
__author_email__ = "vineet.hingorani@telekom.de"
__license__ = "MIT"
__copyright__ = "Copyright 2021, Deutsche Telekom AG"

__spi_version__ = "1.5.1"
